----------------------------
-- Lua3D Graphics Library --
----------------------------

--[[ Copyright and License infomation
	Lua3D Graphics Library module Copyright  20132016 Matthew Cenance. 
	
	The Lua3D Graphics Library module is distributed under the same terms as Lua itself.
	Lua is free software distributed under the terms of the MIT license reproduced here.
	Lua may be used for any purpose, including commercial purposes, at absolutely no cost.
	No paperwork, no royalties, no GNU-like "copyleft" restrictions, either.
	
	Permission is hereby granted, free of charge, to any person obtaining a copy of this
	software and associated documentation files (the "Software"), to deal in the Software
	without restriction, including without limitation the rights to use, copy, modify,
	merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
	permit persons to whom the Software is furnished to do so, subject to the following
	conditions:
	
	The above copyright notice and this permission notice shall be included in all
	copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
	INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
	DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
]]

local Lua3D = {Version = 1.2, VersionStr = "1.2.0"}
Lua3D._VERSION = "Lua3D " .. Lua3D.VersionStr

-- Check the platform we are on
Lua3D.Platform = "Basic Lua" -- The default platform
if sys and draw and turtle then
	Lua3D.Platform = "Touch Lua" -- Those libraries are known to be included with Touch Lua
elseif love then
	Lua3D.Platform = "Love"
elseif type(on) == "table" and type(clipboard) == "table" and type(document) == "table" then
    Lua3D.Platform = "TI-Nspire" 
end

-- Replace the original type with something that will also return the type of an object.
local rawtype = type

_G.rawtype = rawtype
function GetObjectType(object)
	if rawtype(object) ~= "table" then
		return rawtype(object)
	elseif not object.classtype then
		return rawtype(object)
	end
	return object.classtype
end
type = GetObjectType
local type = GetObjectType

-- This function checks for a valid argument type, and throws an error if it is incorrect.
function checkValidArgType(value, types, argnum, level)
	local throwError = true
	if type(types) == "table" then
		for k, allowedType in ipairs(types) do
			if type(value) == allowedType then
				throwError = false
				break
			end
		end
	else
		throwError = type(value) ~= types
	end
	
	-- If we need to throw an error do it
	-- similar to how lua functions error
	if throwError then
		local tab = debug.getinfo(2, "n")
		tab.name = tab.name or "function"
		local errorMessage = "bad argument #"..argnum.." to '"..tab.name.."' ("
		if type(types) == "table" then
			local lastType = types[#types]
			types[#types] = nil
			if #types > 1 then
				errorMessage = errorMessage..table.concat(types, ",")..","
			elseif #types > 0 then
				errorMessage = errorMessage..types[1]
			else
				errorMessage = errorMessage..lastType
			end
			if #types > 0 then
				errorMessage = errorMessage.." or "..lastType
			end
		else
			errorMessage = errorMessage..types
		end
		errorMessage = errorMessage.." expected, got "..type(value)..")"
		error(errorMessage, level or 3)
	end
end

-- Platform independent stuff --
local Support = {
	Rendering = false, 	-- This is set to true if we can draw anything to the screen.
	Shaders = false,	-- This is set to true if we can use shaders.
	Canvas = false,
}

if Lua3D.Platform == "Love" then
	if love.graphics then
		Support.Rendering = true
		if love.graphics.isSupported then
			Support.Shaders = love.graphics.isSupported("shader")
		elseif love.graphics.getSupported then
			local items = love.graphics.getSupported()
			Support.Shaders = items.shader or true
		else
			Support.Shaders = true
		end
	end
elseif Lua3D.Platform == "Touch Lua" then
	Support.Rendering = true
end

-- Platform independent functions --

-- Returns true if this item is supported.
function Lua3D.IsSupported(item)
	return Support[item] or false
end

-- Returns the current time.
function Lua3D.GetTime()
	if Lua3D.Platform == "Love" then
		if love.timer then
			return love.timer.getTime()
		end
	elseif Lua3D.Platform == "Touch Lua" then
		return sys.gettime()
	end
end

-- Add the Lua3D module to the packages that we have loaded, it is required for the Renderer module.
package.loaded["Lua3D"] = Lua3D

-- Detect TI-Nspire platform, and if detected we must do work to set up packaging and a replacement for the io library.
if Lua3D.Platform == "TI-Nspire" then
	
end

-- Require required modules --
local modules = {}
function modules.Color()
	-- Color class --
	local baseColors = {}
	local order = {}
	local Color = {}
	Color.__index = Color -- failed table lookups on the instances should fallback to the class table, to get methods
	Color.classtype = "Color" -- This is the name of the class.

	-- Set some local variables.
	local random, deg = math.random, math.deg
	local setmetatable = setmetatable

	-- This function constructs a new Color object.
	-- Arguments:
	-- numbers red, green, blue, [alpha = 255] are the values to be used in the new Color object.
	-- boolean [raw = false] should be set if the values are already scaled to a range of 0-1.
	-- number hexColor if used as the only argument will act as the hexadecimal value for the color.
	-- Returns:
	-- Color object is the new Color object.
	function Color.new(red, green, blue, alpha, raw)
		-- Check if the arguments are valid
		--[[
		checkValidArgType(red, "number", 1)
		if green then
			checkValidArgType(green, "number", 2)
			checkValidArgType(blue, "number", 3)
			if alpha then
				checkValidArgType(alpha, "number", 4)
			end
		end
		]]
		
		-- If hexColor is the only argument then convert the hex number into the three components.
		if not green then
			local c1, c2, c3 = math.modf(red/(256*256), 256), math.modf(red/256, 256), math.modf(red, 256)
			red, green, blue = c1, c2, c3
		end
		
		-- Create the color object.
		if raw then
			return setmetatable({red = red, green = green, blue = blue, alpha = alpha or 1}, Color)
		else
			return setmetatable({red = red/255, green = green/255, blue = blue/255, alpha = (alpha or 255)/255}, Color)
		end
	end

	-- Creates a new random color.
	-- Arguments:
	-- number [alpha = 255] determines the alpha value of the random color.
	-- Returns:
	-- Color object is the new Color object.
	function Color.random(alpha)
		return Color.new(random()*255, random()*255, random()*255, alpha)
	end

	-- Adds a vector to this color.
	function Color:AddColorToColor(other)
		--[[local output = Color.new(self.red, self.green, self.blue, self.alpha)
		output.red = self.red + vector2.red
		output.green = self.green + vector2.green
		output.blue = self.blue + vector2.blue
		output.alpha = self.alpha + vector2.alpha
		return output]]
		return setmetatable({red = self.red + other.red, green = self.green + other.green, blue = self.blue + other.blue, alpha = self.alpha + other.alpha}, Color)
	end

	-- Subtracts a vector from this vector.
	function Color:SubtractColorFromColor(other)
		--[[local output = Color.new(self.red, self.green, self.blue, self.alpha)
		output.red = self.red - other.red
		output.green = self.green - other.green
		output.blue = self.blue - other.blue
		output.alpha = self.alpha - other.alpha
		return output]]
		return setmetatable({red = self.red - other.red, green = self.green - other.green, blue = self.blue - other.blue, alpha = self.alpha - other.alpha}, Color)
	end

	-- Multiplies a vector by this vector.
	function Color:MultiplyColorByColor(other)
		--[[local output = Color.new(self.red, self.green, self.blue, self.alpha)
		output.red = self.red * other.red
		output.green = self.green * other.green
		output.blue = self.blue * other.blue
		output.alpha = self.alpha * other.alpha
		return output]]
		return setmetatable({red = self.red * other.red, green = self.green * other.green, blue = self.blue * other.blue, alpha = self.alpha * other.alpha}, Color)
	end

	-- Divides a vector by this vector.
	function Color:DivideColorByColor(other)
		--[[local output = Color.new(self.red, self.green, self.blue, self.alpha)
		output.red = self.red / other.red
		output.green = self.green / other.green
		output.blue = self.blue / other.blue
		output.alpha = self.alpha / other.alpha
		return output]]
		return setmetatable({red = self.red / other.red, green = self.green / other.green, blue = self.blue / other.blue, alpha = self.alpha / other.alpha}, Color)
	end

	-- Gets the hue of a color.
	function Color:GetHue()
		local Red, Green, Blue = self.red*255, self.green*255, self.blue*256
		return math.deg(math.atan2(math.sqrt(3)*(Green - Blue), 2*Red - Green - Blue))%360
	end

	-- Returns the raw values for this color.
	function Color:raw()
		return self.red, self.green, self.blue, self.alpha
	end

	-- Set up the Color object's metatable.
	Color.__tostring = function(self)
		if self.red then
			return "Color ("..self.red..", "..self.green..", "..self.blue..", "..self.alpha..")"
		end
		return "Color"
	end

	Color.__add = function(self, other)
		if GetObjectType(other) ~= "Color" then
			if type(other) == "number" then
				return Color.new((self.red*255) + other, (self.green*255) + other, (self.blue*255) + other, (self.alpha*255) + other)
			end
			return self
		end
		return self:AddColorToColor(other)
		-- return setmetatable({red = self.red + other.red, green = self.green + other.green, blue = self.blue + other.blue, alpha = self.alpha + other.alpha}, Color)
	end
	Color.__sub = function(self, other)
		if GetObjectType(other) ~= "Color" then
			if type(other) == "number" then
				return Color.new((self.red*255) - other, (self.green*255) - other, (self.blue*255) - other, (self.alpha*255) - other)
			end
			return self
		end
		return self:SubtractColorFromColor(other)
	end
	Color.__mul = function(self, other)
		if GetObjectType(other) ~= "Color" then
			if type(other) == "number" then
				--return Color.new(self.red*other, self.green*other, self.blue*other, self.alpha, true)
				return setmetatable({red = self.red * other, green = self.green * other, blue = self.blue * other, alpha = self.alpha}, Color)
			end
			return self
		end
		return self:MultiplyColorByColor(other)
		-- return setmetatable({red = self.red * other.red, green = self.green * other.green, blue = self.blue * other.blue, alpha = self.alpha * other.alpha}, Color)
	end
	Color.__div = function(self, other)
		if GetObjectType(other) ~= "Color" then
			if type(other) == "number" then
				return Color.new(self.red/other, self.green/other, self.blue/other, self.alpha, true)
			end
			return self
		end
		return self:DivideColorByColor(other)
	end

	-- Inverts a color.
	Color.__unm = function(self)
		return setmetatable({red = 1 - self.red, green = 1 - self.green, blue = 1 - self.blue, alpha = self.alpha}, Color)
	end

	-- Comparison functions --
	Color.__eq = function(self, other)
		if GetObjectType(other) ~= "Color" then
			return false
		end
		if self.red == other.red and self.green == other.green and self.blue == other.blue then
			return true
		end
		return false
	end
	Color.__lt = function(self, other)
		if GetObjectType(other) ~= "Color" then
			return false
		end
		if self.red < other.red and self.green < other.green and self.blue < other.blue then
			return true
		end
		return false
	end
	Color.__le = function(self, other)
		if GetObjectType(other) ~= "Color" then
			return false
		end
		if self.red <= other.red and self.green <= other.green and self.blue <= other.blue then
			return true
		end
		return false
	end

	return Color
end
function modules.Vector3()
	-- Vector3 class --
	local Vector3 = {}
	Vector3.__index = Vector3 -- failed table lookups on the instances should fallback to the class table, to get methods
	Vector3.classtype = "Vector3"

	local type = type
	local setmetatable = setmetatable
	local rad, cos, sin = math.rad, math.cos, math.sin

	-- This function constructs a new Vector3 object.
	-- Arguments:
	-- numbers x, y, z are the numbers used in the Vector3.
	-- Returns:
	-- Vector3 object is the new Vector3 object.
	function Vector3.new(x, y, z)
		-- Check if the arguments are valid
		--[[
		checkValidArgType(x, "number", 1)
		checkValidArgType(y, "number", 2)
		checkValidArgType(z, "number", 3)
		]]
		
		-- Return the new Vector3 value.
		return setmetatable({x = x, y = y, z = z}, Vector3)
	end

	-- Rotates a vector around the XY plane.
	-- Arguments:
	-- number degrees is the amount of degrees to rotate the Vector3 by.
	-- Returns:
	-- Vector3 output is the new Vector3 object.
	function Vector3:RotateXY(degrees)
		local output = Vector3.new(self.x, self.y, self.z)
		local radians = rad(degrees)
		output.x = cos(radians)*self.x + -sin(radians)*output.y
		output.y = sin(radians)*self.x + cos(radians)*self.y
		return output
	end

	-- Rotates a vector around the YZ plane.
	-- Arguments:
	-- number degrees is the amount of degrees to rotate the Vector3 by.
	-- Returns:
	-- Vector3 output is the new Vector3 object.
	function Vector3:RotateYZ(degrees)
		local output = Vector3.new(self.x, self.y, self.z)
		local radians = rad(degrees)
		output.y = cos(radians)*self.y + -sin(radians)*output.z
		output.z = sin(radians)*self.y + cos(radians)*self.z
		return output
	end

	-- Rotates a vector around the XZ plane.
	-- Arguments:
	-- number degrees is the amount of degrees to rotate the Vector3 by.
	-- Returns:
	-- Vector3 output is the new Vector3 object.
	function Vector3:RotateXZ(degrees)
		local output = Vector3.new(self.x, self.y, self.z)
		local radians = rad(degrees)
		output.x = cos(radians)*self.x + sin(radians)*self.z
		output.z = -sin(radians)*self.x + cos(radians)*self.z
		return output
	end

	-- Returns a Vector3 value that is linearly inpolerated between two Vector3 values.
	-- Arguments:
	-- Vector3 other is the second Vector3 object.
	-- number percent is scaled between zero and one, and determines the position of the output Vector3.
	-- Returns:
	-- Vector3 output is the new Vector3 object.
	function Vector3:Lerp(other, percent)
		return Vector3.new(self.x + (other.x - self.x)*percent, self.y + (other.y - self.y)*percent, self.z + (other.z - self.z)*percent)
	end

	-- Returns a Vector3 object representing the unit of a Vector3 value, scaled so that the magnitude will be equal to one.
	function Vector3:Unit()
		local maximumValue = math.max(math.abs(self.x), math.abs(self.y), math.abs(self.z))
		return setmetatable({x = self.x/maximumValue, y = self.y/maximumValue, z = self.z/maximumValue}, Vector3)
	end

	-- Returns a number representing the magnitude of a Vector3 value.
	function Vector3:Magnitude()
		return self.x + self.y + self.z
	end

	-- Returns three numbers representing the x, y, and z values in this Vector3.
	function Vector3:raw()
		return self.x, self.y, self.z
	end

	-- Set up the metatable --
	Vector3.__tostring = function(self)
		if self.x and self.y and self.z then
			return "Vector3 ("..self.x..", "..self.y..", "..self.z..")"
		end
		return "Vector3"
	end

	-- Mathmatical functions --
	Vector3.__add = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x + other, self.y + other, self.z + other)
			elseif type(other) == "table" then
				return Vector3.new(self.x + (other.x or other.X or other[1] or 0), self.y + (other.y or other.Y or other[2] or 0), self.z + (other.z or other.z or other[3] or 0))
			end
			return self
		end
		return setmetatable({x = self.x + other.x, y = self.y + other.y, z = self.z + other.z}, Vector3)
	end
	Vector3.__sub = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x - other, self.y - other, self.z - other)
			elseif type(other) == "table" then
				return Vector3.new(self.x - (other.x or other.X or other[1] or 0), self.y - (other.y or other.Y or other[2] or 0), self.z - (other.z or other.z or other[3] or 0))
			end
			return self
		end
		return setmetatable({x = self.x - other.x, y = self.y - other.y, z = self.z - other.z}, Vector3)
	end
	Vector3.__mul = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x*other, self.y*other, self.z*other)
			elseif type(other) == "table" then
				return Vector3.new(self.x * (other.x or other.X or other[1] or 1), self.y * (other.y or other.Y or other[2] or 1), self.z * (other.z or other.z or other[3] or 1))
			end
			return self
		end
		return setmetatable({x = self.x * other.x, y = self.y * other.y, z = self.z * other.z}, Vector3)
	end
	Vector3.__div = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x/other, self.y/other, self.z/other)
			elseif type(other) == "table" then
				return Vector3.new(self.x / (other.x or other.X or other[1] or 1), self.y / (other.y or other.Y or other[2] or 1), self.z / (other.z or other.z or other[3] or 1))
			end
			return self
		end
		return setmetatable({x = self.x / other.x, y = self.y / other.y, z = self.z / other.z}, Vector3)
	end
	Vector3.__mod = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x%other, self.y%other, self.z%other)
			elseif type(other) == "table" then
				return Vector3.new(self.x % (other.x or other.X or other[1] or 1), self.y % (other.y or other.Y or other[2] or 1), self.z % (other.z or other.z or other[3] or 1))
			end
			return self
		end
		return setmetatable({x = self.x % other.x, y = self.y % other.y, z = self.z % other.z}, Vector3)
	end
	Vector3.__pow = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			if type(other) == "number" then
				return Vector3.new(self.x^other, self.y^other, self.z^other)
			elseif type(other) == "table" then
				return Vector3.new(self.x ^ (other.x or other.X or other[1] or 1), self.y ^ (other.y or other.Y or other[2] or 1), self.z ^ (other.z or other.z or other[3] or 1))
			end
			return self
		end
		return setmetatable({x = self.x ^ other.x, y = self.y ^ other.y, z = self.z ^ other.z}, Vector3)
	end
	Vector3.__unm = function(self)
		return setmetatable({x = -self.x, y = -self.y, z = -self.z}, Vector3)
	end

	-- Comparison functions --
	Vector3.__eq = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			return false
		end
		if self.x == other.x and self.y == other.y and self.z == other.z then
			return true
		end
		return false
	end
	Vector3.__lt = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			return false
		end
		if self.x < other.x and self.y < other.y and self.z < other.z then
			return true
		end
		return false
	end
	Vector3.__le = function(self, other)
		if GetObjectType(other) ~= "Vector3" then
			return false
		end
		if self.x <= other.x and self.y <= other.y and self.z <= other.z then
			return true
		end
		return false
	end

	return Vector3
end
function modules.Renderer()
	-- Lua3D Renderer class --
	local Lua3D = require("Lua3D")
	local Renderer = {}
	Renderer.__index = Renderer
	Renderer.classtype = "Renderer"
	Renderer.RenderDistance = 128

	local ipairs, pairs = ipairs, pairs
	local setmetatable = setmetatable
	local tostring = tostring
	local type = type

	local cos, sin = math.cos, math.sin
	local rad, deg = math.rad, math.deg
	local max, min = math.max, math.min
	local pi = math.pi

	local insert = table.insert

	local camorx, camory, camorz = 15, 15, 15
	camorx, camory, camorz = rad(camorx), rad(camory), rad(camorz)
	local camvx, camvy, camvz = sin(camorx), cos(camory), camorz
	local cosx, cosy, cosz = cos(camorx), cos(camory), cos(camorz)
	local sinx, siny, sinz = sin(camorx), sin(camory), sin(camorz)

	local zangle = 35 / 180 * pi
	local zcos = cos(zangle)
	local zsin = sin(zangle)

	local width, height = 20, 20
	if Lua3D.Platform == "Love" then
		if love.window then
			width, height = love.window.getWidth(), love.window.getHeight()
		end
	elseif Lua3D.Platform == "Touch Lua" then
		width, height = draw.getport()
	elseif Lua3D.Platform == "TI-Nspire" then
		width, height = platform.window:width(), platform.window:height()
	end

	-- A list of Shaders.
	local shaders = {pixelShaders = {}, vertexShaders = {}}
	if Lua3D.Platform == "Love" then
		if love.graphics then
			shaders.pixelShaders.Default = love.graphics.newShader [[
				vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords)
				{
					vec4 texturecolor = Texel(texture, texture_coords);
					return texturecolor * color;
				}
			]]
			shaders.vertexShaders.Default = love.graphics.newShader [[
				vec4 position(mat4 transform_projection, vec4 vertex_position)
				{
					// The order of operations matters when doing matrix multiplication.
					return transform_projection * vertex_position;
				}
			]]
		end
	end

	-- Sets the maximum render distance of the camera. This will cut off the rendering of polygons farther away than this point.
	-- Arguments:
	-- number value is the new maximum render distance.
	function Renderer:SetCameraRenderDistance(value)
		--checkValidArgType(value, "number", 1)
		self.RenderDistance = max(value, 100)
	end

	-- Sets the Anti-Alias feature of the renderer. By default Anti-Alias is disabled.
	-- Arguments:
	-- number value is the new anti-alias value.
	function Renderer:SetAntiAlias(value)
		--checkValidArgType(value, "boolean", 1)
		self.AntiAliased = value
	end

	-- Sets the camera into Perspective or Graph mode.
	function Renderer:SetRenderMode(value)
		if string.lower(tostring(value)) == "perspective" then
			value = true
		end
		self.PerspectiveMode = value
	end

	-- Sets the camera zoom.
	function Renderer:SetCameraZoom(value)
		--checkValidArgType(value, "number", 1)
		self.CameraZoom = value
	end

	-- Sets the camera's field of view angle.
	function Renderer:SetCameraFOV(value)
		--checkValidArgType(value, "number", 1)
		self.CameraFOV = max(value/45, 1/90)
	end

	-- Sets the position of the camera.
	function Renderer:SetCameraPosition(x, y, z)
		-- Check if the argument type is valid
		--checkValidArgType(x, {"Vector3", "number"}, 1)
		--if type(x) == "number" then
			--checkValidArgType(y, "number", 2)
			--checkValidArgType(z, "number", 3)
		--end
		
		-- Convert a Vector3 to raw variables.
		if type(x) == "Vector3" then
			x, y, z = x:raw()
		end
		
		-- Set the camera position.
		self.CameraPositionX, self.CameraPositionY, self.CameraPositionZ = x, y, z
	end

	-- Sets the camera rotation.
	function Renderer:SetCameraRotation(x, y, z)
		-- Check if the argument type is valid
		--checkValidArgType(x, {"Vector3", "number"}, 1)
		--if type(x) == "number" then
			--checkValidArgType(y, "number", 2)
			--checkValidArgType(z, "number", 3)
		--end
		-- Convert a Vector3 to raw variables.
		if type(x) == "Vector3" then
			x, y, z = x:raw()
		end
		
		-- Set the camera rotation
		camorx, camory, camorz = x, y, z
		camorx, camory, camorz = rad(camorx), rad(camory), rad(camorz)
		self.camorx, self.camory, self.camorz = camorx, camory, camorz
		
		camvx, camvy, camvz = sin(camorx), cos(camory), camorz
		cosx, cosy, cosz = cos(camorx), cos(camory), cos(camorz)
		sinx, siny, sinz = sin(camorx), sin(camory), sin(camorz)
	end

	-- Transforms a 3D position.
	function Renderer:Transform3D(x, y, z)
		-- Check if the argument type is valid
		--checkValidArgType(x, {"Vector3", "number"}, 1)
		--if type(x) == "number" then
			--checkValidArgType(y, "number", 2)
			--checkValidArgType(z, "number", 3)
		--nd
		
		-- Convert a Vector3 to the three numbers.
		if type(x) == "Vector3" then
			x, y, z = x.x, x.y, x.z
		end
		
		-- Transform the value.
		-- Determine the current camera position
		local CameraPositionX, CameraPositionY, CameraPositionZ = self.CameraPositionX, self.CameraPositionY, self.CameraPositionZ
		local fov = self.CameraFOV
		local zoom = self.CameraZoom

		-- Transform the 3D position.
		local xs, ys, zs = (x - CameraPositionX)*zoom, (y - CameraPositionY)*zoom, (z - CameraPositionZ)*zoom
		local ex, ey, ez = 10*fov, 10*fov, 240*fov

		-- Determine if the camera is perspective or not.
		if self.PerspectiveMode then
			-- Use the perspective drawing mode
			local dx = cosy*(sinz*ys + cosz*xs) - siny*zs
			local dy = sinx*(cosy*zs + siny*(sinz*ys + cosz*xs)) + cosx*(cosz*ys - sinz*xs)
			local dz = cosx*(cosy*zs + siny*(sinz*ys + cosz*xs)) - sinx*(cosz*ys - sinz*xs)

			return dx, dy, dz
		else
			-- Use the isometric drawing mode
			ys = -ys
			return (width*.5) + (xs * zoom - zs * zoom * cosz)*.5, (height*.5) + (ys * zoom + zs * zoom * zsin)*.5, zs * zoom + zs * zoom * zsin
		end
	end

	-- Transforms a 3D position into a two dimensional point on the screen, which is used internally to draw objects on to the screen.
	-- Returns:
	-- numbers x, y is the 2D position on the screen to draw the element.
	-- number z is the value used to determine if the object is behind another object.
	local xcenter, ycenter = width*.5, height*.5
	function Renderer:Project3D(x, y, z)
		-- Determine the current camera position
		local CameraPositionX, CameraPositionY, CameraPositionZ = self.CameraPositionX, self.CameraPositionY, self.CameraPositionZ
		local fov = self.CameraFOV
		local zoom = self.CameraZoom

		-- Transform the 3D position.
		local xs, ys, zs = (x - CameraPositionX)*zoom, (y - CameraPositionY)*zoom, (z - CameraPositionZ)*zoom
		local ex, ey, ez = 10*fov, 10*fov, 240*fov

		-- Determine if we should use the perspective view or isometric view.
		if self.PerspectiveMode then
			-- Use the perspective drawing mode
			local dx = cosy*(sinz*ys + cosz*xs) - siny*zs
			local dy = sinx*(cosy*zs + siny*(sinz*ys + cosz*xs)) + cosx*(cosz*ys - sinz*xs)
			local dz = cosx*(cosy*zs + siny*(sinz*ys + cosz*xs)) - sinx*(cosz*ys - sinz*xs)

			-- Calculate position to draw at
			local bx = (ez/dz)*dx - ex
			local by = (ez/dz)*dy - ey
			local bz = ((ez/dz))*(ez - dz)
			--return bx + width*.5, by + height*.5
			--return bx, by
			return bx + (xcenter), by + (ycenter), ez - dz
		else
			-- Use the isometric drawing mode
			ys = -ys
			return (width*.5) + (xs * zoom - zs * zoom * cosz)*.5, (height*.5) + ((ys * zoom + zs * zoom * zsin)*.5), zs * zoom + zs * zoom * zsin
		end
	end

	-- This function queues a three dimensional polygon face to be drawn to the screen. 
	-- To draw queued 3D elements call Renderer:DrawItemsToScreen.
	function Renderer:DrawPolygon(point1, point2, point3, point4, color, texture)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"Polygon", 0, point1, point2, point3, point4, color, texture}
	end

	-- Queues a three dimensional polygon face outline to be drawn to the screen.
	function Renderer:DrawPolygonOutline(point1, point2, point3, point4, color)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"PolygonOutline", 0, point1, point2, point3, point4, color}
	end

	-- Queues a three dimensional triangle to be drawn by the renderer.
	function Renderer:DrawTriangle(point1, point2, point3, color, texture)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"Triangle", 0, point1, point2, point3, color, texture}
	end

	-- Queues a three dimensional triangle to be drawn by the renderer.
	function Renderer:DrawTriangleOutline(point1, point2, point3, color)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"TriangleOutline", 0, point1, point2, point3, color}
	end

	-- Queues a three dimensional line to be drawn by the renderer.
	function Renderer:DrawLine(point1, point2, color, width)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"Line", width or 2, point1, point2, color}
	end

	-- Queues a sprite to be drawn on the screen.
	function Renderer:DrawSprite(image, position, size)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"Sprite", image, position, size}
	end

	-- Queues a string of text to be drawn on the screen.
	function Renderer:DrawString(text, position, size, color)
		self.ItemsToRender[#self.ItemsToRender + 1] = {"String", text, position, size, color}
	end

	-- The mesh used when rendering items to the screen.
	local vertices = {
		{
			-- Top-left corner
			0, 0, -- position of the vertex
			0, 0, -- texture coordinate at the vertex position
			255, 0, 0, -- color of the vertex
		},
		{
			-- Top-right corner
			0, 0,
			1, 0, -- texture coordinates are in the range of [0, 1]
			0, 255, 0
		},
		{
			-- Bottom-right corner
			0, 0,
			1, 1,
			0, 0, 255
		},
		{
			-- Bottom-left corner
			0, 0,
			0, 1,
			255, 255, 0
		},
	}
	local textureMesh
	if Lua3D.Platform == "Love" then
		textureMesh = love.graphics.newMesh(vertices, nil, "fan")
	end

	-- This function draws queued 3D elements to the screen.
	-- Returns:
	-- table objectsDrawn is a table containing the amount of which objects that were drawn to the screen.
	function Renderer:DrawItemsToScreen(gc)
		if Lua3D.Platform == "Love" then
			width, height = love.window.getWidth(), love.window.getHeight()
		end
		
		local startTime = Lua3D.GetTime()
		local minz, maxz = self.CameraFOV*240, self.CameraFOV*self.RenderDistance
		local objectsQueued = {}
		local objectsDrawn = {}
		local order = {}
		local Project3D = self.Project3D
		
		local p1, p2, p3, p4
		
		-- Prepare the queued elements to be drawn.
		for k, v in ipairs(self.ItemsToRender) do
			objectsQueued[v[1]] = (objectsQueued[v[1]] or 0) + 1
			if v[1] == "Polygon" or v[1] == "PolygonOutline" then
				-- Draw a flat polygon in 3D space.
				-- Get the ordered points of this
				-- polygon.
				p1, p2, p3, p4 = v[3], v[4], v[5], v[6]
				-- Transform the coordinates of this
				-- polygon.
				local x1, y1, zvalue = Project3D(self, p1:raw())
				local x2, y2, zvalue2 = Project3D(self, p2:raw())
				local x3, y3, zvalue3 = Project3D(self, p3:raw())
				local x4, y4, zvalue4 = Project3D(self, p4:raw())
				local zvalue1 = zvalue
				-- Determine the Z value of this
				-- polygon.
				local zvalueLow = min(zvalue, zvalue2, zvalue3, zvalue4)
				local zvalueHigh = max(zvalue, zvalue2, zvalue3, zvalue4)
				zvalue = zvalueLow + ((zvalueHigh - zvalueLow)*.1)
				-- If the zvalue is in range, draw it.
				if zvalueLow - minz > 0 and zvalueHigh - minz <= maxz then
					if not order[zvalue] then
						order[zvalue] = {}
					end
					local col1, col2, col3, col4 = v[7], v[7], v[7], v[7]
					if self.FogEnabled then
						col1 = v[7]*max(min((maxz - zvalue1)/(maxz / 2), 1), 0)
						col2 = v[7]*max(min((maxz - zvalue2 - 100)/(maxz / 2), 1), 0)
						col3 = v[7]*max(min((maxz - zvalue3 - 100)/(maxz / 2), 1), 0)
						col4 = v[7]*max(min((maxz - zvalue4 - 100)/(maxz / 2), 1), 0)
					end
					order[zvalue][#order[zvalue] + 1] = {v[1], x1, y1, x2, y2, x3, y3, x4, y4, col1, col2, col3, col4, v[8], zvalue, zvalue2, zvalue3, zvalue4}
				end
			elseif v[1] == "Triangle" or v[1] == "TriangleOutline" then
				-- Draw a triangle in 3D space.
				p1, p2, p3 = v[3], v[4], v[5]
				local x1, y1, zvalue = Project3D(self, p1:raw())
				local x2, y2, zvalue2 = Project3D(self, p2:raw())
				local x3, y3, zvalue3 = Project3D(self, p3:raw())
				local zvalue = zvalue - minz
				local zvalue1 = zvalue
				local zvalue2 = zvalue2 - minz
				local zvalue3 = zvalue3 - minz
				local zvalueLow = min(zvalue, zvalue2, zvalue3)
				local zvalueHigh = max(zvalue, zvalue2, zvalue3)
				zvalue = zvalueLow + ((zvalueHigh - zvalueLow)*.1)
				if zvalueLow > 0 and zvalueHigh <= maxz then
					if not order[zvalue] then
						order[zvalue] = {}
					end
					local col1, col2, col3 = v[6], v[6], v[6]
					if self.FogEnabled then
						col1 = v[6]*max(min((maxz - zvalue1)/(maxz / 2), 1), 0)
						col2 = v[6]*max(min((maxz - zvalue2)/(maxz / 2), 1), 0)
						col3 = v[6]*max(min((maxz - zvalue3)/(maxz / 2), 1), 0)
					end
					order[zvalue][#order[zvalue] + 1] = {v[1], x1, y1, x2, y2, x3, y3, col1, col2, col3, v[7]}
				end
			elseif v[1] == "Sprite" then
				-- Draw a two dimensional sprite
				-- in 3D space.
				p1 = v[3]
				local x1, y1, zvalue = Project3D(self, p1:raw())
				if zvalue > minz and zvalue <= maxz then
					if not order[zvalue] then
						order[zvalue] = {}
					end
					order[zvalue][#order[zvalue] + 1] = {"Sprite", v[2], x1, y1, ((1/((zvalue - minz)*.2 + 1))*v[4])/(v[2]:getHeight())}
				end
			elseif v[1] == "String" then
				-- Draw a string in 3D space.
				p1 = v[3]
				local x1, y1, zvalue = Project3D(self, p1:raw())
				if zvalue > minz*2 then
					if not order[zvalue] then
						order[zvalue] = {}
					end
					order[zvalue][#order[zvalue] + 1] = {"String", v[2], x1, y1, v[4], v[5]}
				end
			elseif v[1] == "Line" then
				-- Draw a line in 3D space.
				p1, p2 = v[3], v[4]
				local x1, y1, zvalue = Project3D(self, p1:raw())
				local x2, y2, zvalue2 = Project3D(self, p2:raw())
				local zvalueLow = min(zvalue, zvalue2)
				local zvalueHigh = max(zvalue, zvalue2)
				zvalue = zvalueLow + ((zvalueHigh - zvalueLow)*.1)
				if zvalueLow > minz then
					if not order[zvalue] then
						order[zvalue] = {}
					end
					order[zvalue][#order[zvalue] + 1] = {"Line", x1, y1, x2, y2, v[5], v[2]}
				end
			end
		end
		
		local stepOneDoneTime = Lua3D.GetTime()

		-- Draw queued elements to screen.
		local dOrder = {}
		local sOrder = {}
		-- We must create a sorted table, so we can draw items to the screen
		-- in the correct order so that objects further away draw first.
		-- We can't simply use pairs because then objects would be drawn
		-- in the order that they were queued, which is not correct.
		local insertAtEnd
		for k, v in pairs(order) do
			if #sOrder == 0 then
				sOrder[1] = k
			else
				insertAtEnd = true
				
				-- Search through the table and insert it.
				for k2, v2 in ipairs(sOrder) do
					if k > v2 then
						insert(sOrder, k2, k)
						insertAtEnd = false
						break
					end
				end
				
				-- If we did not already insert it into the table, insert it to the end.
				if insertAtEnd then
					sOrder[#sOrder + 1] = k
				end
			end
		end
		
		local stepTwoDoneTime = Lua3D.GetTime()
		
		-- Sort through the sOrder table and insert it to the dOrder table.
		for k, v in ipairs(sOrder) do
			dOrder[k] = order[v]
		end
		
		local stepThreeDoneTime = Lua3D.GetTime()
		
		-- Draw the elements to the screen.
		if Lua3D.Platform == "Love" then
			for k, v in ipairs(dOrder) do
				for k2, v2 in ipairs(v) do
					objectsDrawn[v2[1]] = (objectsDrawn[v2[1]] or 0) + 1
					if v2[1] == "Polygon" then
						if math.min(v2[2], v2[4], v2[6], v2[8]) > -256 and math.max(v2[2], v2[4], v2[6], v2[8]) < width + 256 then
						if math.min(v2[3], v2[5], v2[7], v2[9]) > -256 and math.max(v2[3], v2[5], v2[7], v2[9]) < height + 256 then
						local col1, col2, col3, col4 = v2[10], v2[11], v2[12], v2[13]
						local vertex1 = {v2[2], v2[3], 0, 0, col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255}
						local vertex2 = {v2[4], v2[5], 1, 0, col2.red*255, col2.green*255, col2.blue*255, col2.alpha*255}
						local vertex3 = {v2[6], v2[7], 1, 1, col3.red*255, col3.green*255, col3.blue*255, col3.alpha*255}
						local vertex4 = {v2[8], v2[9], 0, 1, col4.red*255, col4.green*255, col4.blue*255, col4.alpha*255}
						textureMesh:setTexture( v2[14] )
						love.graphics.setColor(col1.red*255, col1.blue*255, col1.green*255, col1.alpha*255)
						textureMesh:setVertices( {vertex1, vertex2, vertex3, vertex4} )
						love.graphics.draw(textureMesh)
						end
						end
					elseif v2[1] == "PolygonOutline" then
						local col = v2[10]
						love.graphics.setColor(col.red*255, col.blue*255, col.green*255, col.alpha*255)
						love.graphics.polygon("line", v2[2], v2[3], v2[4], v2[5], v2[6], v2[7])
						love.graphics.polygon("line", v2[2], v2[3], v2[6], v2[7], v2[8], v2[9])
					elseif v2[1] == "Triangle" then
						local col1, col2, col3 = v2[8], v2[9], v2[10]
						local vertex1 = {v2[2], v2[3], 0, 0, col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255}
						local vertex2 = {v2[4], v2[5], 1, 0, col2.red*255, col2.green*255, col2.blue*255, col2.alpha*255}
						local vertex3 = {v2[6], v2[7], 1, 1, col3.red*255, col3.green*255, col3.blue*255, col3.alpha*255}
						textureMesh:setTexture( v2[11] )
						love.graphics.setColor(col1.red*255, col1.blue*255, col1.green*255, col1.alpha*255)
						textureMesh:setVertices( {vertex1, vertex2, vertex3} )
						love.graphics.draw(textureMesh)
					elseif v2[1] == "TriangleOutline" then
						local col = v2[8]
						love.graphics.setColor(col.red*255, col.blue*255, col.green*255, col.alpha*255)
						love.graphics.polygon("line", v2[2], v2[3], v2[4], v2[5], v2[6], v2[7])
					elseif v2[1] == "Sprite" then
						love.graphics.setColor(255, 255, 255, 255)
						love.graphics.draw(v2[2], v2[3], v2[4], -camorz, v2[5], v2[5], v2[2]:getWidth()*.5, v2[2]:getHeight()*.5)
					elseif v2[1] == "String" then
						love.graphics.setColor(v2[6].red*255, v2[6].blue*255, v2[6].green*255, v2[6].alpha*255)
						local sizeX, sizeY = #v2, 1
						love.graphics.printf(v2[2], v2[3] - (sizeX*.5), v2[4], 200, "center")
					elseif v2[1] == "Line" then
						local col = v2[6]
						love.graphics.setColor(col.red*255, col.blue*255, col.green*255, col.alpha*255)
						love.graphics.line(v2[2], v2[3], v2[4], v2[5])
					end
				end
			end
		elseif Lua3D.Platform == "TI-Nspire" then
			for k, v in ipairs(dOrder) do
				for k2, v2 in ipairs(v) do
					objectsDrawn[v2[1]] = (objectsDrawn[v2[1]] or 0) + 1
					if v2[1] == "Polygon" then
						if math.min(v2[2], v2[4], v2[6], v2[8]) > -256 and math.max(v2[2], v2[4], v2[6], v2[8]) < width + 256 then
							if math.min(v2[3], v2[5], v2[7], v2[9]) > -256 and math.max(v2[3], v2[5], v2[7], v2[9]) < height + 256 then
								local col1, col2, col3, col4 = v2[10], v2[11], v2[12], v2[13]
								--local vertex1 = {v2[2], v2[3], 0, 0, col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255}
								--local vertex2 = {v2[4], v2[5], 1, 0, col2.red*255, col2.green*255, col2.blue*255, col2.alpha*255}
								--local vertex3 = {v2[6], v2[7], 1, 1, col3.red*255, col3.green*255, col3.blue*255, col3.alpha*255}
								--local vertex4 = {v2[8], v2[9], 0, 1, col4.red*255, col4.green*255, col4.blue*255, col4.alpha*255}
								--textureMesh:setTexture( v2[14] )
								gc:setColorRGB(col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255)
								--textureMesh:setVertices( {vertex1, vertex2, vertex3, vertex4} )
								gc:fillPolygon({v2[2], v2[3], v2[4], v2[5], v2[6], v2[7], v2[8], v2[9]})
							end
						end
					elseif v2[1] == "PolygonOutline" then
						local col = v2[10]
						gc:setColorRGB(col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255)
						love.graphics.polygon("line", v2[2], v2[3], v2[4], v2[5], v2[6], v2[7])
						love.graphics.polygon("line", v2[2], v2[3], v2[6], v2[7], v2[8], v2[9])
					elseif v2[1] == "Triangle" then
						local col1, col2, col3 = v2[8], v2[9], v2[10]
						--[[local vertex1 = {v2[2], v2[3], 0, 0, col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255}
						local vertex2 = {v2[4], v2[5], 1, 0, col2.red*255, col2.green*255, col2.blue*255, col2.alpha*255}
						local vertex3 = {v2[6], v2[7], 1, 1, col3.red*255, col3.green*255, col3.blue*255, col3.alpha*255}]]
						--textureMesh:setTexture( v2[11] )
						gc:setColorRGB(col1.red*255, col1.green*255, col1.blue*255, col1.alpha*255)
						--textureMesh:setVertices( {vertex1, vertex2, vertex3} )
						pcall(gc.fillPolygon, gc, {v2[2], v2[3], v2[4], v2[5], v2[6], v2[7]})
					elseif v2[1] == "TriangleOutline" then
						local col = v2[8]
						gc:setColorRGB(col.red*255, col.green*255, col.blue*255, col.alpha*255)
						pcall(gc.drawPolyLine, gc, {v2[2], v2[3], v2[4], v2[5], v2[6], v2[7], v2[2], v2[3]})
					elseif v2[1] == "Sprite" then
						love.graphics.draw(v2[2], v2[3], v2[4], -camorz, v2[5], v2[5], v2[2]:getWidth()*.5, v2[2]:getHeight()*.5)
					elseif v2[1] == "String" then
						local col = v2[6]
						gc:setColorRGB(col.red*255, col.green*255, col.blue*255, col.alpha*255)
						gc:drawString(v2[2], v2[3], v2[4])
					elseif v2[1] == "Line" then
						local col = v2[6]
						gc:setColorRGB(col.red*255, col.blue*255, col.green*255, col.alpha*255)
						gc:drawLine(v2[2], v2[3], v2[4], v2[5])
					end
				end
			end
		end
		
		local doneTime = Lua3D.GetTime()
		
		-- Draw the FPS, camera position, rotation, zoom, field of view, and maximum render distance.
		local leftHeight = 0
		if Lua3D.Platform == "Love" then
			if self.DrawRenderInformation then
				love.graphics.setColor(255, 255, 255, 255)
				love.graphics.print("<Render Info>", 10, leftHeight + 10)
				love.graphics.print("Current Camera Position: ", 10, leftHeight + 30)
				love.graphics.print(tostring("X: "..self.CameraPositionX.." Y: "..self.CameraPositionY.." Z: "..self.CameraPositionZ), 10, leftHeight + 50)
				love.graphics.print("Current Camera Rotation: ", 10, leftHeight + 80)
				love.graphics.print(tostring("X: "..camorx.." Y: "..camory.." Z: "..camorz), 10, leftHeight + 100)
				love.graphics.print("Current Zoom: "..self.CameraZoom, 10, leftHeight + 130)
				love.graphics.print("Current Fov: "..self.CameraFOV, 10, leftHeight + 150)
				love.graphics.print("Render Distance: "..self.RenderDistance.." units", 10, leftHeight + 170)
				love.graphics.print("Memory Used: "..math.ceil(collectgarbage("count")/1024).." megabytes", 10, leftHeight + 190)
				leftHeight = leftHeight + 210
			end
			
			if self.DrawRenderProfile then
				love.graphics.setColor(255, 255, 255, 255)
				love.graphics.print("<Render Profiler>", 10, leftHeight + 10)
				love.graphics.print("Current FPS: "..tostring(love.timer.getFPS( )), 10, leftHeight + 30)
				love.graphics.print("Time taken to Draw: "..math.floor((doneTime - startTime)*1000000).." microseconds", 10, leftHeight + 50)
				love.graphics.print("Step 1 Transform Elements: "..math.floor((stepOneDoneTime - startTime)*1000000).." microseconds", 10, leftHeight + 80)
				love.graphics.print("Step 2 Sort Elements 1: "..math.floor((stepTwoDoneTime - stepOneDoneTime)*1000000).." microseconds", 10, leftHeight + 100)
				love.graphics.print("Step 3 Sort Elements 2: "..math.floor((stepThreeDoneTime - stepTwoDoneTime)*1000000).." microseconds", 10, leftHeight + 120)
				love.graphics.print("Step 4 Draw Elements: "..math.floor((doneTime - stepThreeDoneTime)*1000000).." microseconds", 10, leftHeight + 140)
			end
			
			-- Draw the objects drawn last frame.
			if self.DrawObjectsDrawn then
				love.graphics.setColor(255, 255, 255, 255)
				love.graphics.print("<Objects Drawn>", 400, 10)
				local line = 0
				for key, value in pairs(objectsDrawn) do
					line = line + 1
					love.graphics.print(key.." x"..value, 400, 10 + line*20)
				end
			end
		end
		
		self.ItemsToRender = {}
		return objectsQueued
	end

	-- This function constructs a new Renderer object.
	-- Returns:
	-- Render object is the new Renderer object.
	Renderer.new = function(xCam, yCam, zCam)
		local self = setmetatable({}, Renderer)
		self.CameraPositionX, self.CameraPositionY, self.CameraPositionZ = xCam or 0, yCam or 0, zCam or 0 -- The camera position.
		self.CameraFOV = 1 -- The field of view angle.
		self.CameraZoom = .25 -- How large items will be drawn on the screen.
		self.PerspectiveMode = true -- Determines if we will use the perspective view or the isometric view.
		self.FogEnabled = true -- Determines if items will fade away when they are getting close to the edges of the render distance.
		self.DrawObjectsDrawn = true -- Determines if we will draw the information about objects drawn in the last frame.
		self.DrawRenderInformation = true -- Determines if we will draw the information about the renderer.
		self.DrawRenderProfile = true
		self.ItemsToRender = {}
		self:SetCameraRotation(xRot or 0, yRot or 0, zRot or 0)
		return self
	end

	return Renderer
end
function modules.bit32()
	--[[

	LUA MODULE

		bit.numberlua - Bitwise operations implemented in pure Lua as numbers,
			with Lua 5.2 'bit32' and (LuaJIT) LuaBitOp 'bit' compatibility interfaces.

	SYNOPSIS

		local bit = require 'bit.numberlua'
		print(bit.band(0xff00ff00, 0x00ff00ff)) --> 0xffffffff
		
		-- Interface providing strong Lua 5.2 'bit32' compatibility
		local bit32 = require 'bit.numberlua'.bit32
		assert(bit32.band(-1) == 0xffffffff)
		
		-- Interface providing strong (LuaJIT) LuaBitOp 'bit' compatibility
		local bit = require 'bit.numberlua'.bit
		assert(bit.tobit(0xffffffff) == -1)
		
	DESCRIPTION
		
		This library implements bitwise operations entirely in Lua.
		This module is typically intended if for some reasons you don't want
		to or cannot	install a popular C based bit library like BitOp 'bit' [1]
		(which comes pre-installed with LuaJIT) or 'bit32' (which comes
		pre-installed with Lua 5.2) but want a similar interface.
		
		This modules represents bit arrays as non-negative Lua numbers. [1]
		It can represent 32-bit bit arrays when Lua is compiled
		with lua_Number as double-precision IEEE 754 floating point.

		The module is nearly the most efficient it can be but may be a few times
		slower than the C based bit libraries and is orders or magnitude
		slower than LuaJIT bit operations, which compile to native code.	Therefore,
		this library is inferior in performane to the other modules.

		The `xor` function in this module is based partly on Roberto Ierusalimschy's
		post in http://lua-users.org/lists/lua-l/2002-09/msg00134.html .
		
		The included BIT.bit32 and BIT.bit sublibraries aims to provide 100%
		compatibility with the Lua 5.2 "bit32" and (LuaJIT) LuaBitOp "bit" library.
		This compatbility is at the cost of some efficiency since inputted
		numbers are normalized and more general forms (e.g. multi-argument
		bitwise operators) are supported.
		
	STATUS

		WARNING: Not all corner cases have been tested and documented.
		Some attempt was made to make these similar to the Lua 5.2 [2]
		and LuaJit BitOp [3] libraries, but this is not fully tested and there
		are currently some differences.	Addressing these differences may
		be improved in the future but it is not yet fully determined how to
		resolve these differences.
		
		The BIT.bit32 library passes the Lua 5.2 test suite (bitwise.lua)
		http://www.lua.org/tests/5.2/ .	The BIT.bit library passes the LuaBitOp
		test suite (bittest.lua).	However, these have not been tested on
		platforms with Lua compiled with 32-bit integer numbers.

	API

		BIT.tobit(x) --> z
		
			Similar to function in BitOp.
			
		BIT.tohex(x, n)
		
			Similar to function in BitOp.
		
		BIT.band(x, y) --> z
		
			Similar to function in Lua 5.2 and BitOp but requires two arguments.
		
		BIT.bor(x, y) --> z
		
			Similar to function in Lua 5.2 and BitOp but requires two arguments.

		BIT.bxor(x, y) --> z
		
			Similar to function in Lua 5.2 and BitOp but requires two arguments.
		
		BIT.bnot(x) --> z
		
			Similar to function in Lua 5.2 and BitOp.

		BIT.lshift(x, disp) --> z
		
			Similar to function in Lua 5.2 (warning: BitOp uses unsigned lower 5 bits of shift),
		
		BIT.rshift(x, disp) --> z
		
			Similar to function in Lua 5.2 (warning: BitOp uses unsigned lower 5 bits of shift),

		BIT.extract(x, field [, width]) --> z
		
			Similar to function in Lua 5.2.
		
		BIT.replace(x, v, field, width) --> z
		
			Similar to function in Lua 5.2.
		
		BIT.bswap(x) --> z
		
			Similar to function in Lua 5.2.

		BIT.rrotate(x, disp) --> z
		BIT.ror(x, disp) --> z
		
			Similar to function in Lua 5.2 and BitOp.

		BIT.lrotate(x, disp) --> z
		BIT.rol(x, disp) --> z

			Similar to function in Lua 5.2 and BitOp.
		
		BIT.arshift
		
			Similar to function in Lua 5.2 and BitOp.
			
		BIT.btest
		
			Similar to function in Lua 5.2 with requires two arguments.

		BIT.bit32
		
			This table contains functions that aim to provide 100% compatibility
			with the Lua 5.2 "bit32" library.
			
			bit32.arshift (x, disp) --> z
			bit32.band (...) --> z
			bit32.bnot (x) --> z
			bit32.bor (...) --> z
			bit32.btest (...) --> true | false
			bit32.bxor (...) --> z
			bit32.extract (x, field [, width]) --> z
			bit32.replace (x, v, field [, width]) --> z
			bit32.lrotate (x, disp) --> z
			bit32.lshift (x, disp) --> z
			bit32.rrotate (x, disp) --> z
			bit32.rshift (x, disp) --> z

		BIT.bit
		
			This table contains functions that aim to provide 100% compatibility
			with the LuaBitOp "bit" library (from LuaJIT).
			
			bit.tobit(x) --> y
			bit.tohex(x [,n]) --> y
			bit.bnot(x) --> y
			bit.bor(x1 [,x2...]) --> y
			bit.band(x1 [,x2...]) --> y
			bit.bxor(x1 [,x2...]) --> y
			bit.lshift(x, n) --> y
			bit.rshift(x, n) --> y
			bit.arshift(x, n) --> y
			bit.rol(x, n) --> y
			bit.ror(x, n) --> y
			bit.bswap(x) --> y
			
	DEPENDENCIES

		None (other than Lua 5.1 or 5.2).
			
	DOWNLOAD/INSTALLATION

		If using LuaRocks:
			luarocks install lua-bit-numberlua

		Otherwise, download <https://github.com/davidm/lua-bit-numberlua/zipball/master>.
		Alternately, if using git:
			git clone git://github.com/davidm/lua-bit-numberlua.git
			cd lua-bit-numberlua
		Optionally unpack:
			./util.mk
		or unpack and install in LuaRocks:
			./util.mk install 

	REFERENCES

		[1] http://lua-users.org/wiki/FloatingPoint
		[2] http://www.lua.org/manual/5.2/
		[3] http://bitop.luajit.org/
		
	LICENSE

		(c) 2008-2011 David Manura.	Licensed under the same terms as Lua (MIT).

		Permission is hereby granted, free of charge, to any person obtaining a copy
		of this software and associated documentation files (the "Software"), to deal
		in the Software without restriction, including without limitation the rights
		to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
		copies of the Software, and to permit persons to whom the Software is
		furnished to do so, subject to the following conditions:

		The above copyright notice and this permission notice shall be included in
		all copies or substantial portions of the Software.

		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
		IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
		FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.	IN NO EVENT SHALL THE
		AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
		LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
		OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
		THE SOFTWARE.
		(end license)

	--]]

	local M = {_TYPE='module', _NAME='bit.numberlua', _VERSION='0.3.1.20120131'}

	local floor = math.floor

	local MOD = 2^32
	local MODM = MOD-1

	local function memoize(f)
		local mt = {}
		local t = setmetatable({}, mt)
		function mt:__index(k)
			local v = f(k); t[k] = v
			return v
		end
		return t
	end

	local function make_bitop_uncached(t, m)
		local function bitop(a, b)
			local res,p = 0,1
			while a ~= 0 and b ~= 0 do
				local am, bm = a%m, b%m
				res = res + t[am][bm]*p
				a = (a - am) / m
				b = (b - bm) / m
				p = p*m
			end
			res = res + (a+b)*p
			return res
		end
		return bitop
	end

	local function make_bitop(t)
		local op1 = make_bitop_uncached(t,2^1)
		local op2 = memoize(function(a)
			return memoize(function(b)
				return op1(a, b)
			end)
		end)
		return make_bitop_uncached(op2, 2^(t.n or 1))
	end

	-- ok?	probably not if running on a 32-bit int Lua number type platform
	function M.tobit(x)
		return x % 2^32
	end

	M.bxor = make_bitop {[0]={[0]=0,[1]=1},[1]={[0]=1,[1]=0}, n=4}
	local bxor = M.bxor

	function M.bnot(a)	 return MODM - a end
	local bnot = M.bnot

	function M.band(a,b) return ((a+b) - bxor(a,b))/2 end
	local band = M.band

	function M.bor(a,b)	return MODM - band(MODM - a, MODM - b) end
	local bor = M.bor

	local lshift, rshift -- forward declare

	function M.rshift(a,disp) -- Lua5.2 insipred
		if disp < 0 then return lshift(a,-disp) end
		return floor(a % 2^32 / 2^disp)
	end
	rshift = M.rshift

	function M.lshift(a,disp) -- Lua5.2 inspired
		if disp < 0 then return rshift(a,-disp) end 
		return (a * 2^disp) % 2^32
	end
	lshift = M.lshift

	function M.tohex(x, n) -- BitOp style
		n = n or 8
		local up
		if n <= 0 then
			if n == 0 then return '' end
			up = true
			n = - n
		end
		x = band(x, 16^n-1)
		return ('%0'..n..(up and 'X' or 'x')):format(x)
	end
	local tohex = M.tohex

	function M.extract(n, field, width) -- Lua5.2 inspired
		width = width or 1
		return band(rshift(n, field), 2^width-1)
	end
	local extract = M.extract

	function M.replace(n, v, field, width) -- Lua5.2 inspired
		width = width or 1
		local mask1 = 2^width-1
		v = band(v, mask1) -- required by spec?
		local mask = bnot(lshift(mask1, field))
		return band(n, mask) + lshift(v, field)
	end
	local replace = M.replace

	function M.bswap(x)	-- BitOp style
		local a = band(x, 0xff); x = rshift(x, 8)
		local b = band(x, 0xff); x = rshift(x, 8)
		local c = band(x, 0xff); x = rshift(x, 8)
		local d = band(x, 0xff)
		return lshift(lshift(lshift(a, 8) + b, 8) + c, 8) + d
	end
	local bswap = M.bswap

	function M.rrotate(x, disp)	-- Lua5.2 inspired
		disp = disp % 32
		local low = band(x, 2^disp-1)
		return rshift(x, disp) + lshift(low, 32-disp)
	end
	local rrotate = M.rrotate

	function M.lrotate(x, disp)	-- Lua5.2 inspired
		return rrotate(x, -disp)
	end
	local lrotate = M.lrotate

	M.rol = M.lrotate	-- LuaOp inspired
	M.ror = M.rrotate	-- LuaOp insipred


	function M.arshift(x, disp) -- Lua5.2 inspired
		local z = rshift(x, disp)
		if x >= 0x80000000 then z = z + lshift(2^disp-1, 32-disp) end
		return z
	end
	local arshift = M.arshift

	function M.btest(x, y) -- Lua5.2 inspired
		return band(x, y) ~= 0
	end

	--
	-- Start Lua 5.2 "bit32" compat section.
	--

	M.bit32 = {} -- Lua 5.2 'bit32' compatibility


	local function bit32_bnot(x)
		return (-1 - x) % MOD
	end
	M.bit32.bnot = bit32_bnot

	local function bit32_bxor(a, b, c, ...)
		local z
		if b then
			a = a % MOD
			b = b % MOD
			z = bxor(a, b)
			if c then
				z = bit32_bxor(z, c, ...)
			end
			return z
		elseif a then
			return a % MOD
		else
			return 0
		end
	end
	M.bit32.bxor = bit32_bxor

	local function bit32_band(a, b, c, ...)
		local z
		if b then
			a = a % MOD
			b = b % MOD
			z = ((a+b) - bxor(a,b)) / 2
			if c then
				z = bit32_band(z, c, ...)
			end
			return z
		elseif a then
			return a % MOD
		else
			return MODM
		end
	end
	M.bit32.band = bit32_band

	local function bit32_bor(a, b, c, ...)
		local z
		if b then
			a = a % MOD
			b = b % MOD
			z = MODM - band(MODM - a, MODM - b)
			if c then
				z = bit32_bor(z, c, ...)
			end
			return z
		elseif a then
			return a % MOD
		else
			return 0
		end
	end
	M.bit32.bor = bit32_bor

	function M.bit32.btest(...)
		return bit32_band(...) ~= 0
	end

	function M.bit32.lrotate(x, disp)
		return lrotate(x % MOD, disp)
	end

	function M.bit32.rrotate(x, disp)
		return rrotate(x % MOD, disp)
	end

	function M.bit32.lshift(x,disp)
		if disp > 31 or disp < -31 then return 0 end
		return lshift(x % MOD, disp)
	end

	function M.bit32.rshift(x,disp)
		if disp > 31 or disp < -31 then return 0 end
		return rshift(x % MOD, disp)
	end

	function M.bit32.arshift(x,disp)
		x = x % MOD
		if disp >= 0 then
			if disp > 31 then
				return (x >= 0x80000000) and MODM or 0
			else
				local z = rshift(x, disp)
				if x >= 0x80000000 then z = z + lshift(2^disp-1, 32-disp) end
				return z
			end
		else
			return lshift(x, -disp)
		end
	end

	function M.bit32.extract(x, field, ...)
		local width = ... or 1
		if field < 0 or field > 31 or width < 0 or field+width > 32 then error 'out of range' end
		x = x % MOD
		return extract(x, field, ...)
	end

	function M.bit32.replace(x, v, field, ...)
		local width = ... or 1
		if field < 0 or field > 31 or width < 0 or field+width > 32 then error 'out of range' end
		x = x % MOD
		v = v % MOD
		return replace(x, v, field, ...)
	end


	--
	-- Start LuaBitOp "bit" compat section.
	--

	M.bit = {} -- LuaBitOp "bit" compatibility

	function M.bit.tobit(x)
		x = x % MOD
		if x >= 0x80000000 then x = x - MOD end
		return x
	end
	local bit_tobit = M.bit.tobit

	function M.bit.tohex(x, ...)
		return tohex(x % MOD, ...)
	end

	function M.bit.bnot(x)
		return bit_tobit(bnot(x % MOD))
	end

	local function bit_bor(a, b, c, ...)
		if c then
			return bit_bor(bit_bor(a, b), c, ...)
		elseif b then
			return bit_tobit(bor(a % MOD, b % MOD))
		else
			return bit_tobit(a)
		end
	end
	M.bit.bor = bit_bor

	local function bit_band(a, b, c, ...)
		if c then
			return bit_band(bit_band(a, b), c, ...)
		elseif b then
			return bit_tobit(band(a % MOD, b % MOD))
		else
			return bit_tobit(a)
		end
	end
	M.bit.band = bit_band

	local function bit_bxor(a, b, c, ...)
		if c then
			return bit_bxor(bit_bxor(a, b), c, ...)
		elseif b then
			return bit_tobit(bxor(a % MOD, b % MOD))
		else
			return bit_tobit(a)
		end
	end
	M.bit.bxor = bit_bxor

	function M.bit.lshift(x, n)
		return bit_tobit(lshift(x % MOD, n % 32))
	end

	function M.bit.rshift(x, n)
		return bit_tobit(rshift(x % MOD, n % 32))
	end

	function M.bit.arshift(x, n)
		return bit_tobit(arshift(x % MOD, n % 32))
	end

	function M.bit.rol(x, n)
		return bit_tobit(lrotate(x % MOD, n % 32))
	end

	function M.bit.ror(x, n)
		return bit_tobit(rrotate(x % MOD, n % 32))
	end

	function M.bit.bswap(x)
		return bit_tobit(bswap(x % MOD))
	end

	return M
end
function modules.voxelPolygonizer()
	local bit32 = modules.bit32()

	local edgeTable = {
		0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
		0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
		0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
		0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
		0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
		0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
		0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
		0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
		0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
		0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
		0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
		0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
		0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
		0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
		0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
		0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
		0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
		0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
		0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
		0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
		0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
		0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
		0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
		0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
		0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
		0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
		0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
		0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
		0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
		0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
		0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
		0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0   
	};
	local triTable = {
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
		{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
		{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
		{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
		{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
		{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
		{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
		{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
		{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
		{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
		{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
		{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
		{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
		{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
		{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
		{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
		{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
		{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
		{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
		{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
		{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
		{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
		{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
		{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
		{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
		{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
		{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
		{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
		{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
		{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
		{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
		{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
		{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
		{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
		{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
		{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
		{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
		{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
		{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
		{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
		{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
		{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
		{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
		{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
		{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
		{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
		{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
		{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
		{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
		{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
		{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
		{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
		{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
		{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
		{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
		{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
		{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
		{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
		{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
		{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
		{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
		{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
		{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
		{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
		{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
		{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
		{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
		{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
		{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
		{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
		{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
		{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
		{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
		{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
		{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
		{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
		{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
		{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
		{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
		{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
		{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
		{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
		{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
		{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
		{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
		{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
		{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
		{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
		{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
		{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
		{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
		{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
		{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
		{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
		{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
		{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
		{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
		{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
		{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
		{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
		{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
		{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
		{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
		{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
		{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
		{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
		{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
		{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
		{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
		{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
		{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
		{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
		{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
		{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
		{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
		{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
		{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
		{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
		{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
		{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
		{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
		{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
		{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
		{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
		{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
		{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
		{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
		{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
		{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
		{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
		{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
		{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
		{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
		{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
		{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
		{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
		{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
		{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
		{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
		{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
		{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
		{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
		{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
		{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
		{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
		{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
		{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
		{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
		{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
		{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
		{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
		{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
		{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
		{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
		{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
		{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
	}

	local bor = bit32.bor
	local band = bit32.band
	local abs = math.abs

	local voxelPolygonizer = {}

	--[[
	   Linearly interpolate the position where an isosurface cuts
	   an edge between two vertices, each with their own scalar value
	]]
	local VertexInterp = function(isolevel,p1,p2,valp1,valp2)
		--double isolevel;
		--XYZ p1,p2;
		--double valp1,valp2;
		local mu --double mu;
		local p = {x = 0, y = 0, z = 0}--XYZ p;

		if (abs(isolevel-valp1) < 0.00001) then
			return p1
		end
		if (abs(isolevel-valp2) < 0.00001) then
			return p2
		end
		if (abs(valp1-valp2) < 0.00001) then
			return p1
		end
		mu = (isolevel - valp1) / (valp2 - valp1);
		p.x = p1.x + mu * (p2.x - p1.x);
		p.y = p1.y + mu * (p2.y - p1.y);
		p.z = p1.z + mu * (p2.z - p1.z);

		return p
	end

	--[[
		Given a grid cell and an isolevel, calculate the triangular
		facets required to represent the isosurface through the cell.
		Return the number of triangular facets, the array "triangles"
		will be loaded up with the vertices at most 5 triangular facets.
		0 will be returned if the grid cell is either totally above
		of totally below the isolevel.
	]]
	-- Gridcell, number, triangle
	function voxelPolygonizer.Polygonise(grid, isolevel, triangles)
		local i, ntriang
		local cubeindex = 0
		local vertlist = {}

		--[[
			Determine the index into the edge table which
			tells us which vertices are inside of the surface
		]]
		if (grid.val[0] < isolevel) then cubeindex = bor(cubeindex, 1) end
		if (grid.val[1] < isolevel) then cubeindex = bor(cubeindex, 2) end
		if (grid.val[2] < isolevel) then cubeindex = bor(cubeindex, 4) end
		if (grid.val[3] < isolevel) then cubeindex = bor(cubeindex, 8) end
		if (grid.val[4] < isolevel) then cubeindex = bor(cubeindex, 16) end
		if (grid.val[5] < isolevel) then cubeindex = bor(cubeindex, 32) end
		if (grid.val[6] < isolevel) then cubeindex = bor(cubeindex, 64) end
		if (grid.val[7] < isolevel) then cubeindex = bor(cubeindex, 128) end
		
		cubeindex = cubeindex + 1
		
		-- Cube is entirely in/out of the surface
		if (edgeTable[cubeindex] == 0) then
			return 0
		end

		-- Find the vertices where the surface intersects the cube
		if (band(edgeTable[cubeindex], 1) > 0) then
			vertlist[0] = VertexInterp(isolevel,grid.p[0],grid.p[1],grid.val[0],grid.val[1]);
		end 
		if (band(edgeTable[cubeindex],  2) > 0) then
			vertlist[1] = VertexInterp(isolevel,grid.p[1],grid.p[2],grid.val[1],grid.val[2]);
		end 
		if (band(edgeTable[cubeindex],  4) > 0) then
			vertlist[2] = VertexInterp(isolevel,grid.p[2],grid.p[3],grid.val[2],grid.val[3]);
		end 
		if (band(edgeTable[cubeindex],  8) > 0) then
			vertlist[3] = VertexInterp(isolevel,grid.p[3],grid.p[0],grid.val[3],grid.val[0]);
		end 
		if (band(edgeTable[cubeindex],  16) > 0) then
			vertlist[4] = VertexInterp(isolevel,grid.p[4],grid.p[5],grid.val[4],grid.val[5]);
		end 
		if (band(edgeTable[cubeindex],  32) > 0) then
			vertlist[5] = VertexInterp(isolevel,grid.p[5],grid.p[6],grid.val[5],grid.val[6]);
		end 
		if (band(edgeTable[cubeindex],  64) > 0) then
			vertlist[6] = VertexInterp(isolevel,grid.p[6],grid.p[7],grid.val[6],grid.val[7]);
		end 
		if (band(edgeTable[cubeindex],  128) > 0) then
			vertlist[7] = VertexInterp(isolevel,grid.p[7],grid.p[4],grid.val[7],grid.val[4]);
		end 
		if (band(edgeTable[cubeindex],  256) > 0) then
			vertlist[8] = VertexInterp(isolevel,grid.p[0],grid.p[4],grid.val[0],grid.val[4]);
		end 
		if (band(edgeTable[cubeindex],  512) > 0) then
			vertlist[9] = VertexInterp(isolevel,grid.p[1],grid.p[5],grid.val[1],grid.val[5]);
		end 
		if (band(edgeTable[cubeindex],  1024) > 0) then
			vertlist[10] = VertexInterp(isolevel,grid.p[2],grid.p[6],grid.val[2],grid.val[6]);
		end 
		if (band(edgeTable[cubeindex],  2048) > 0) then
			vertlist[11] = VertexInterp(isolevel,grid.p[3],grid.p[7],grid.val[3],grid.val[7]);
		end

		-- Create the triangle
		ntriang = 0
		i = 0
		--for (i=0, triTable[cubeindex][i]!=-1, i+=3) do
		while triTable[cubeindex][i] ~= -1 do
			triangles[ntriang] = {p = {}}
			triangles[ntriang].p[0] = vertlist[triTable[cubeindex][i+1]];
			triangles[ntriang].p[1] = vertlist[triTable[cubeindex][i+2]];
			triangles[ntriang].p[2] = vertlist[triTable[cubeindex][i+3]];
			ntriang = ntriang + 1;
			i = i + 3
			if i > 90 then break end
		end
		
		return ntriang
	end

	return voxelPolygonizer
end
local chunkRenderRange = 1
function modules.SmoothTerrain()
	local voxelPolygonizer = modules.voxelPolygonizer()
	local Polygonise = voxelPolygonizer.Polygonise

	---------------------------
	-- Smooth Terrain object --
	---------------------------
	local Terrain = {}
	Terrain.__index = Terrain

	local chunkSize = 8
	
	local setmetatable = setmetatable
	local pairs, ipairs = pairs, ipairs

	-- Creates a new terrain object.
	function Terrain.new()
		local self = setmetatable({}, Terrain)
		
		self.Materials = {}
		self.MaterialData = {}
		self.OccupancyData = setmetatable({}, {__index = function()
			return setmetatable({}, {__index = function()
				return setmetatable({}, {__index = function() return 0 end})
			end})
		end})
		
		self.DrawingMeshes = {}
		self.MeshesToReset = {}
		
		return self
	end

	-- Regenerates the drawing mesh.
	function Terrain:RegenerateChunkMesh(xChunk, yChunk, zChunk, detail)
		if self.ChunksToGenerateThisFrame <= 0 then
			return false
		end
		self.ChunksToGenerateThisFrame = self.ChunksToGenerateThisFrame - 1
		local terrainTriangles = {}

		-- Generate the terrainTriangles table
		local triangles, grid = {}, {val = {}}
		if detail <= 1 then
			for y = yChunk*chunkSize, (yChunk + 1)*chunkSize do
				for x = xChunk*chunkSize, (xChunk + 1)*chunkSize do
					for z = zChunk*chunkSize, (zChunk + 1)*chunkSize do
						grid.p = {
							{x = 0 + x, y = 0 + y, z = 0 + z},
							{x = 1 + x, y = 0 + y, z = 0 + z},
							{x = 1 + x, y = 0 + y, z = 1 + z},
							{x = 0 + x, y = 0 + y, z = 1 + z},
							{x = 0 + x, y = 1 + y, z = 0 + z},
							{x = 1 + x, y = 1 + y, z = 0 + z},
							{x = 1 + x, y = 1 + y, z = 1 + z},
							{x = 0 + x, y = 1 + y, z = 1 + z},
						}
						for vertex = 0, 7 do
							grid.p[vertex] = grid.p[vertex + 1]
							grid.val[vertex] = self.OccupancyData[grid.p[vertex].y][grid.p[vertex].x][grid.p[vertex].z]
						end
						for triangleId = 0, Polygonise(grid, detail, triangles) do
							terrainTriangles[#terrainTriangles + 1] = triangles[triangleId]
						end
					end
				end
			end
		else
			detail = math.ceil(detail)
			for y = yChunk*chunkSize/detail, (yChunk + 1)*chunkSize/detail do
				for x = xChunk*chunkSize/detail, (xChunk + 1)*chunkSize/detail do
					for z = zChunk*chunkSize/detail, (zChunk + 1)*chunkSize/detail do
						grid.p = {
							{x = 0 + x, y = 0 + y, z = 0 + z},
							{x = 1 + x, y = 0 + y, z = 0 + z},
							{x = 1 + x, y = 0 + y, z = 1 + z},
							{x = 0 + x, y = 0 + y, z = 1 + z},
							{x = 0 + x, y = 1 + y, z = 0 + z},
							{x = 1 + x, y = 1 + y, z = 0 + z},
							{x = 1 + x, y = 1 + y, z = 1 + z},
							{x = 0 + x, y = 1 + y, z = 1 + z},
						}
						for a = 1, 8 do
							grid.p[a].x = grid.p[a].x*detail
							grid.p[a].y = grid.p[a].y*detail
							grid.p[a].z = grid.p[a].z*detail
						end
						for vertex = 0, 7 do
							grid.p[vertex] = grid.p[vertex + 1]
							grid.val[vertex] = self.OccupancyData[grid.p[vertex].y][grid.p[vertex].x][grid.p[vertex].z]
						end
						for triangleId = 0, Polygonise(grid, 1, triangles) do
							terrainTriangles[#terrainTriangles + 1] = triangles[triangleId]
						end
					end
				end
			end
		end

		local green = Color.new(0, 255, 0)
		local finalTriangles = {}
		for key, value in pairs(terrainTriangles) do
			local vrt1 = value.p[0]
			local vrt2 = value.p[1]
			local vrt3 = value.p[2]
			if vrt1 and vrt2 and vrt3 then
				finalTriangles[#finalTriangles + 1] = {
					Vector3.new(vrt3.x, vrt3.y, vrt3.z),
					Vector3.new(vrt2.x, vrt2.y, vrt2.z),
					Vector3.new(vrt1.x, vrt1.y, vrt1.z),
					green,
					TextureTriangle
				}
			end
		end
		
		-- Find duplicates
		local duplicateTriangles = {}
		local nonDuplicateTriangles = {}
		for key, value in pairs(finalTriangles) do
			local duplicateTriangle = false
			for key2, value2 in pairs(nonDuplicateTriangles) do
				if value2[1] == value[1] and value2[2] == value[2] and value2[3] == value[3] and (key ~= key2) then
					duplicateTriangles[#duplicateTriangles + 1] = key
					duplicateTriangle = true
					break
				end
			end
			if not duplicateTriangle then
				nonDuplicateTriangles[#nonDuplicateTriangles + 1] = value
			end
		end
		
		for key, value in pairs(duplicateTriangles) do
			table.remove(finalTriangles, value - (key - 1))
		end
		
		-- Add required parent tables
		if not self.DrawingMeshes[xChunk] then
			self.DrawingMeshes[xChunk] = {}
		end
		if not self.DrawingMeshes[xChunk][yChunk] then
			self.DrawingMeshes[xChunk][yChunk] = {}
		end
		if not self.DrawingMeshes[xChunk][yChunk][zChunk] then
			self.DrawingMeshes[xChunk][yChunk][zChunk] = {}
		end
		self.DrawingMeshes[xChunk][yChunk][zChunk][detail] = finalTriangles
		return true
	end

	-- Draws the mesh
	function Terrain:DrawChunk(renderer, xChunk, yChunk, zChunk, detail)
		if detail > 16 then
			detail = 32
		elseif detail > 8 then
			detail = 16
		elseif detail > 4 then
			detail = 8
		elseif detail > 2 then
			detail = 4
		elseif detail > 1 then
			detail = 2
		end
		local chunkExists = false
		if self.DrawingMeshes[xChunk] then
			if self.DrawingMeshes[xChunk][yChunk] then
				if self.DrawingMeshes[xChunk][yChunk][zChunk] then
					if self.DrawingMeshes[xChunk][yChunk][zChunk][detail] then
						chunkExists = true
					end
				end
			end
		end
		if not chunkExists then
			if not self:RegenerateChunkMesh(xChunk, yChunk, zChunk, detail) then
				return false
			end
		end
		for key, value in pairs(self.DrawingMeshes[xChunk][yChunk][zChunk][detail]) do
			renderer:DrawTriangle(
				value[1],
				value[2],
				value[3],
				value[4],
				value[5]
			)
			local blue = Color.new(0, 0, 255)
			renderer:DrawTriangleOutline(
				value[1],
				value[2],
				value[3],
				value[4]*.5,
				value[5]
			)
		end
	end

	-- Draws the terrain.
	function Terrain:Draw(renderer)
		self.ChunksToGenerateThisFrame = 4
		local detailLevel = .05
		local posX = math.floor(renderer.CameraPositionX/chunkSize)
		local posY = math.floor(renderer.CameraPositionY/chunkSize)
		local posZ = math.floor(renderer.CameraPositionZ/chunkSize)
		if #self.MeshesToReset > 0 then
			for key, value in pairs(self.MeshesToReset) do
				if self.DrawingMeshes[value.x] then
					if self.DrawingMeshes[value.x][value.y] then
						self.DrawingMeshes[value.x][value.y][value.z] = nil
					end
				end
			end
			self.MeshesToReset = {}
			--collectgarbage()
		end
		for x = posX - chunkRenderRange, posX + chunkRenderRange do
			for y = posY - chunkRenderRange, posY + chunkRenderRange do
				for z = posZ - chunkRenderRange, posZ + chunkRenderRange do
					self:DrawChunk(renderer, x, y, z, math.abs(x - posX) + math.abs(y - posY) + math.abs(z - posZ) < 4 and detailLevel or 4)
				end
			end
		end
	end

	return Terrain
end
Color = modules.Color()
Vector3 = modules.Vector3()
Renderer = modules.Renderer()
Terrain = modules.SmoothTerrain()

Lua3D.Color = Color
Lua3D.Vector3 = Vector3
Lua3D.Renderer = Renderer

local terrain = Terrain.new()
local renderer = Renderer.new()

local height = 48
local size = 23
local terrainMode = "Hilly"
if terrainMode == "Flat" then
	for y = 0, height do
		terrain.OccupancyData[y] = setmetatable({}, {__index = function()
				return setmetatable({}, {__index = function() return 0 end})
			end})
		for x = -size, size do
			terrain.OccupancyData[y][x] = setmetatable({}, {__index = function() return 0 end, __newindex = function(self, key, value) if key <= size and key >= -size then rawset(self, key, value) end end})
			for z = -size, size do
				terrain.OccupancyData[y][x][z] = y <= 0 and 1 or 0
			end
		end
	end
	renderer.CameraPositionY = 2
else
	for y = 0, height do
		terrain.OccupancyData[y] = setmetatable({}, {__index = function()
				return setmetatable({}, {__index = function() return 0 end})
			end})
		for x = -size, size do
			terrain.OccupancyData[y][x] = setmetatable({}, {__index = function() return 0 end, __newindex = function(self, key, value) if key <= size and key >= -size then rawset(self, key, value) end end})
			for z = -size, size do
				terrain.OccupancyData[y][x][z] = math.max(math.min(math.sin(x*.2)*2 + math.sin((z - .5)*.2)*2 - (y - 8), 1), 0)
			end
		end
	end
	renderer.CameraPositionY = 9
end

local ShowControls = true
local ControlText = {
	"Press H to hide controls",
	"Press Q to go higher or E to go lower",
	"Press Enter to add terrain",
	"Press Delete to remove terrain",
	"Press R to raise terrain",
	"Press L to lower terrain",
}

local cameraRotation = Vector3.new(0, 0, 0)
local white = Color.new(255, 255, 255)
renderer:SetCameraRotation(cameraRotation.x, cameraRotation.y, cameraRotation.z)
platform.window:setBackgroundColor(0x1717ff)
function on.paint(gc, x, y, xwidth, xheight)
	gc:setColorRGB(80, 180, 255)
	gc:fillRect(x, y, xwidth, xheight)
	terrain:Draw(renderer)
	
	-- Draw the building area
	renderer:DrawLine(Vector3.new(size, 1, size), Vector3.new(0, 1, size), white)
	renderer:DrawLine(Vector3.new(-size, 1, size), Vector3.new(0, 1, size), white)
	renderer:DrawLine(Vector3.new(size, 1, -size), Vector3.new(0, 1, -size), white)
	renderer:DrawLine(Vector3.new(-size, 1, -size), Vector3.new(0, 1, -size), white)
	renderer:DrawLine(Vector3.new(size, 1, size), Vector3.new(size, 1, 0), white)
	renderer:DrawLine(Vector3.new(-size, 1, size), Vector3.new(-size, 1, 0), white)
	renderer:DrawLine(Vector3.new(size, 1, -size), Vector3.new(size, 1, -0), white)
	renderer:DrawLine(Vector3.new(-size, 1, -size), Vector3.new(-size, 1, -0), white)
	renderer:DrawLine(Vector3.new(size, height, size), Vector3.new(0, height, size), white)
	renderer:DrawLine(Vector3.new(-size, height, size), Vector3.new(0, height, size), white)
	renderer:DrawLine(Vector3.new(size, height, -size), Vector3.new(0, height, -size), white)
	renderer:DrawLine(Vector3.new(-size, height, -size), Vector3.new(0, height, -size), white)
	renderer:DrawLine(Vector3.new(size, height, size), Vector3.new(size, height, 0), white)
	renderer:DrawLine(Vector3.new(-size, height, size), Vector3.new(-size, height, 0), white)
	renderer:DrawLine(Vector3.new(size, height, -size), Vector3.new(size, height, -0), white)
	renderer:DrawLine(Vector3.new(-size, height, -size), Vector3.new(-size, height, -0), white)
	renderer:DrawLine(Vector3.new(-size, 1, -size), Vector3.new(-size, height, -size), white)
	renderer:DrawLine(Vector3.new(size, 1, -size), Vector3.new(size, height, -size), white)
	renderer:DrawLine(Vector3.new(size, 1, size), Vector3.new(size, height, size), white)
	renderer:DrawLine(Vector3.new(-size, 1, size), Vector3.new(-size, height, size), white)
	
	renderer:DrawString("You may only build in the area marked by the white lines", Vector3.new(0, 4, 0), white)
	
	renderer:DrawItemsToScreen(gc)
	
	if ShowControls then
		for key = 1, #ControlText do
			gc:setColorRGB(0, 0, 0)
			gc:drawString(ControlText[key], 5, 5 + (key - 1)*20)
			gc:setColorRGB(255, 255, 255)
			gc:drawString(ControlText[key], 4, 4 + (key - 1)*20)
		end
	end
end
function on.timer()
	--renderer:SetCameraRotation(math.random()*360, math.random()*360, math.random()*360)
	if terrain.ChunksToGenerateThisFrame < 4 then
		platform.window:invalidate()
	end
end
function on.arrowKey(key)
	if key == "up" or key == "down" then
		local move = Vector3.new(0, 0, key == "up" and -1 or 1)
		move = move:RotateXZ(cameraRotation.y)
		renderer.CameraPositionX = renderer.CameraPositionX + move.x
		renderer.CameraPositionY = renderer.CameraPositionY + move.y
		renderer.CameraPositionZ = renderer.CameraPositionZ + move.z
	elseif key == "left" or key == "right" then
		cameraRotation.y = cameraRotation.y + (key == "left" and -2 or 2)
		renderer:SetCameraRotation(cameraRotation.x, cameraRotation.y, cameraRotation.z)
	end
end
local editRegion = Vector3.new(2, 2, 2)
local editPower = .1
local click = function(mode)
	for x = -editRegion.x*.5, editRegion.x*.5 do
		for y = -editRegion.y*.5, editRegion.y*.5 do
			for z = 1, editRegion.z do
				local position = Vector3.new(x, y, -z)
				position = position:RotateXZ(math.deg(renderer.camory))
				position = position:RotateYZ(math.deg(renderer.camorx))
				position = position:RotateXY(math.deg(renderer.camorz))
				position = position + Vector3.new(renderer.CameraPositionX, renderer.CameraPositionY, renderer.CameraPositionZ)
				position.x = math.floor(position.x + .5)
				position.y = math.floor(position.y + .5)
				position.z = math.floor(position.z + .5)
				
				local occupancy = (terrain.OccupancyData[position.y][position.x][position.z] or 0)
				if mode == "erase" then
					occupancy = math.max(occupancy - (editPower), 0)
				elseif mode == "fill" then
					occupancy = math.min(occupancy + (editPower), 1)
				elseif mode == "raise" then
					occupancy = math.min(occupancy + ((terrain.OccupancyData[position.y - 1][position.x][position.z] or 0) > .8 and editPower or 0), 1)
				elseif mode == "lower" then
					occupancy = math.min(occupancy - ((terrain.OccupancyData[position.y + 1][position.x][position.z] or 0) < .1 and editPower or 0), 1)
				end
				terrain.OccupancyData[position.y][position.x][position.z] = occupancy
				
				terrain.MeshesToReset[#terrain.MeshesToReset + 1] = {
					x = math.floor(position.x/8),
					y = math.floor(position.y/8),
					z = math.floor(position.z/8)
				}
				
				platform.window:invalidate()
				
			end
		end
	end
end
function on.enterKey() click("fill") end
function on.deleteKey() click("erase") end
function on.charIn(char)
	if char == "q" then
		renderer.CameraPositionY = renderer.CameraPositionY + 1
		platform.window:invalidate()
	elseif char == "e" then
		renderer.CameraPositionY = renderer.CameraPositionY - 1
		platform.window:invalidate()
	elseif char == "r" then
		click("raise")
	elseif char == "l" then
		click("lower")
	elseif char == "s" then
		click("smooth")
	elseif char == "h" then
		ShowControls = not ShowControls
	elseif char == "i" then
		chunkRenderRange = chunkRenderRange + 1
	elseif char == "o" then
		chunkRenderRange = math.max(chunkRenderRange - 1, 0)
	elseif char == "p" then
		chunkRenderRange = 1
	end
end
timer.start(.1)